"""
A teeny-weeny synchronous/asynchronous HTTP library. This is not intended to
rival `httpx` or `aiohttp`, but rather to provide a simple interface and be
small. Use it if you just need to make the occasional HTTP request and don't
want to depend on a large library.
"""

import asyncio
import json as json_module
import ssl
import typing as t
import urllib.error
import urllib.request

# Re-export JSONDecodeError, since it can be raised by this module
from json import JSONDecodeError as JSONDecodeError

__all__ = [
    "HttpError",
    "HttpResponse",
    "JSONDecodeError",
    "request",
    "request_sync",
]


HttpMethod = t.Literal[
    "get",
    "GET",
    "head",
    "HEAD",
    "options",
    "OPTIONS",
    "trace",
    "TRACE",
    "put",
    "PUT",
    "delete",
    "DELETE",
    "post",
    "POST",
    "patch",
    "PATCH",
    "connect",
    "CONNECT",
]


HTTP_METHOD_VALUES = t.get_args(HttpMethod)


class HttpError(Exception):
    """
    Raised when an HTTP request fails.

    The status code is `None` if the error wasn't caused by HTTP, but some other
    reason like a network error.


    ## Attributes

    `method`: The HTTP method used in the failed request

    `url`: The target URL of the failed request

    `message`: A human-readable error message

    `status_code`: The HTTP status code, if applicable. This will be `None` if
        the error was not caused by HTTP itself, but rather an external issue
        such as missing permissions or a network error.
    """

    def __init__(
        self,
        *,
        method: HttpMethod,
        url: str,
        status_code: int | None,
        message: str,
    ) -> None:
        super().__init__(
            method,
            url,
            status_code,
            message,
        )

    @property
    def method(self) -> str:
        return self.args[0]

    @property
    def url(self) -> str:
        return self.args[1]

    @property
    def status_code(self) -> int | None:
        return self.args[2]

    @property
    def message(self) -> str:
        return self.args[3]


class HttpResponse:
    """
    Represents the response received by a server after making an HTTP request.


    ## Attributes

    `status_code`: The HTTP status code of the response

    `headers`: A dictionary of headers received in the response. These are all
        lowercase so you can access them without worrying about casing.
    """

    def __init__(
        self,
        *,
        status_code: int,
        headers: dict[str, str],
        content: bytes,
    ) -> None:
        self.status_code = status_code
        self.headers = headers
        self._content = content

    def read(self) -> bytes:
        """
        Returns the response body as bytes.
        """
        return self._content

    def json(self) -> t.Any:
        """
        Returns the response body as a JSON object.


        ## Raises

        `json.JSONDecodeError`: If the response body is not valid JSON. To
            simplify your error handling, this will also consider invalid UTF-8
            to be invalid JSON. (Which is in line with the JSON spec, which
            requires all JSON documents to be encoded in UTF-8.)
        """

        try:
            return json_module.loads(self._content)
        except UnicodeDecodeError:
            raise json_module.JSONDecodeError(
                "The response body is not valid UTF-8",
                "",
                0,
            )


def request_sync(
    method: HttpMethod,
    url: str,
    *,
    content: str | bytes | None = None,
    json: dict[str, t.Any] | None = None,
    headers: dict[str, str] | None = None,
    verify_ssl: bool = True,
) -> HttpResponse:
    """
    Makes a synchronous HTTP request with the specified parameters, returning
    the response. Prefer using the asynchronous `request` function instead.


    ## Parameters

    `method`: What sort of HTTP request to make

    `url`: The URL to make the request to

    `content`: The body of the request, as either a string or bytes. May be
        `None` to make a request without a body.

    `json`: Convenience parameter to specify the body of the request as a JSON
        object. This will also change the `content-type` header to
        `"application/json"`. If this parameter is specified, `content` must be
        `None`.

    `headers`: Additional headers to include in the request. Any headers
        provided here will override the default headers.

    `verify_ssl`: Whether to verify the SSL certificate of the server.


    ## Raises

    `ValueError`: If the `method` is not one of `"get"` or `"post"`

    `HttpError`: If the request fails for any reason. This includes external
        errors such as network issues, as well as any non-200 status codes.
    """
    # Verify the method
    if method not in HTTP_METHOD_VALUES:
        raise ValueError("Invalid method")

    # Prepare a set of default headers
    all_headers = {
        "user-agent": "rio.arequests/0.1",
        "content-type": "application/octet-stream",
    }

    # Prepare the request
    req = urllib.request.Request(
        url,
        method=method.upper(),
    )

    # If a JSON object was provided, use it as content and also set the
    # content-type header
    if json:
        if content is not None:
            raise ValueError("Cannot specify both `content` and `json`")

        content = json_module.dumps(json)
        all_headers["content-type"] = "application/json"

    # If content was provided add it to the request
    if content:
        if isinstance(content, str):
            content = content.encode("utf-8")

        req.data = content

    # Add the headers
    #
    # User-provided headers override the default headers. Take care to do this
    # late so that the logic above (e.g. JSON) can modify the default headers.
    if headers:
        for header, value in headers.items():
            all_headers[header.lower()] = value

    for key, value in all_headers.items():
        req.add_header(key, value)

    # Prepare the SSL context
    #
    # The default context verifies SSL. If that is not desired, override it.
    if verify_ssl:
        ssl_context = None
    else:
        ssl_context = ssl.create_default_context()
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE

    # Make the request
    try:
        with urllib.request.urlopen(req, context=ssl_context) as response:
            # Check the status code
            if response.status >= 300:
                raise HttpError(
                    method=method,
                    url=url,
                    status_code=response.status,
                    message=response.reason,
                )

            # Epic success!
            return HttpResponse(
                status_code=response.status,
                headers={
                    key.lower(): value for key, value in response.getheaders()
                },
                content=response.read(),
            )

    except urllib.error.HTTPError as e:
        raise HttpError(
            method=method,
            url=url,
            status_code=e.code,
            message=e.reason,
        ) from None

    except urllib.error.URLError as e:
        raise HttpError(
            method=method,
            url=url,
            status_code=None,
            message=str(e.reason),
        ) from None


async def request(
    method: HttpMethod,
    url: str,
    *,
    content: bytes | None = None,
    json: dict[str, t.Any] | None = None,
    headers: dict[str, str] | None = None,
    verify_ssl: bool = True,
) -> HttpResponse:
    """
    Makes an asynchronous HTTP request with the specified parameters, returning
    the response.


    ## Parameters

    `method`: What sort of HTTP request to make

    `url`: The URL to make the request to

    `content`: The body of the request, as either a string or bytes. May be
        `None` to make a request without a body.

    `json`: Convenience parameter to specify the body of the request as a JSON
        object. If this is specified, `content` must be `None`.

    `headers`: Additional headers to include in the request. Any headers
        provided here will override the default headers.

    `verify_ssl`: Whether to verify the SSL certificate of the server.


    ## Raises

    `ValueError`: If the `method` is not one of `"get"` or `"post"`

    `HttpError`: If the request fails for any reason. This includes external
        errors such as network issues, as well as any non-200 status codes.
    """

    return await asyncio.to_thread(
        request_sync,
        method,
        url,
        content=content,
        json=json,
        headers=headers,
        verify_ssl=verify_ssl,
    )



import functools
import textwrap
from pathlib import Path

import pytest
from pyfakefs.fake_filesystem import FakeFilesystem, PatchMode
from pyfakefs.fake_filesystem_unittest import Patcher

import rio
from rio.cli.run_project.app_loading import load_user_app
from rio.project_config import RioProjectConfig


# Per default, importing from the fake file system doesn't work. So we must
# define a custom fixture which enables that.
@pytest.fixture
def fs():
    with Patcher(patch_open_code=PatchMode.AUTO) as p:
        yield p.fs


def create_project(
    fs: FakeFilesystem,
    file_hierarchy: str,
    *,
    app_file: str,
    main_module: str,
) -> RioProjectConfig:
    FILE_CONTENTS_BY_NAME = {
        "foo_page.py": """
import rio


@rio.page(url_segment='foo-was-loaded-correctly')
class FooPage(rio.Component):
    def build(self):
        return rio.Text('foo')
"""
    }

    file_hierarchy = textwrap.dedent(file_hierarchy).strip()

    directories_by_indent = {0: Path()}
    rio_toml_location: Path | None = None

    for line in file_hierarchy.splitlines():
        file_name = line.lstrip()
        indent = (len(line) - len(file_name)) // 4
        directory = directories_by_indent[indent]

        # Create the file/folder
        if file_name.endswith("/"):
            new_directory = directory / file_name.rstrip("/")
            directories_by_indent[indent + 1] = new_directory

            # new_directory.mkdir()
            fs.create_dir(new_directory)
        else:
            new_file = directory / file_name

            fs.create_file(
                new_file, contents=FILE_CONTENTS_BY_NAME.get(file_name, "")
            )

            if file_name == "rio.toml":
                rio_toml_location = new_file

    (directories_by_indent[0] / app_file).write_text("""
import rio

app = rio.App(build=rio.Spacer)
""")

    assert rio_toml_location is not None

    return RioProjectConfig(
        file_path=rio_toml_location.absolute(),
        toml_dict={
            "app": {
                "app-type": "website",
                "main-module": main_module,
            }
        },
        dirty_keys=set(),
    )


def assert_page_was_loaded_correctly(
    page: rio.ComponentPage | rio.Redirect, expected_url_segment: str
):
    if page.url_segment == expected_url_segment:
        return

    assert isinstance(page, rio.ComponentPage)
    assert isinstance(page.build, functools.partial)

    error_summary, error_details = page.build.args
    pytest.fail(f"Page {page.name!r} couldn't be loaded: {error_details}")


def test_project_file(fs: FakeFilesystem):
    config = create_project(
        fs,
        """
        my-project/
            my_app.py
            pages/
                foo_page.py
            rio.toml
        """,
        app_file="my-project/my_app.py",
        main_module="my_app",
    )
    app = load_user_app(config).app

    assert app.name == "My App"
    assert app.assets_dir == Path("my-project/assets").absolute()
    assert_page_was_loaded_correctly(app.pages[0], "foo-was-loaded-correctly")


def test_src_folder(fs: FakeFilesystem):
    config = create_project(
        fs,
        """
        my-project/
            src/
                assets/
                my_app.py
                pages/
                    foo_page.py
            rio.toml
        """,
        app_file="my-project/src/my_app.py",
        main_module="my_app",
    )
    app = load_user_app(config).app

    assert app.name == "My App"
    assert app.assets_dir == Path("my-project/src/assets").absolute()
    assert_page_was_loaded_correctly(app.pages[0], "foo-was-loaded-correctly")


def test_simple_project_dir(fs: FakeFilesystem):
    config = create_project(
        fs,
        """
        my-project/
            my_project/
                __init__.py
                assets/
                pages/
                    foo_page.py
            rio.toml
        """,
        app_file="my-project/my_project/__init__.py",
        main_module="my_project",
    )
    app = load_user_app(config).app

    assert app.name == "My Project"
    assert app.assets_dir == Path("my-project/my_project/assets").absolute()
    assert_page_was_loaded_correctly(app.pages[0], "foo-was-loaded-correctly")


def test_submodule(fs: FakeFilesystem):
    config = create_project(
        fs,
        """
        my-project/
            my_project/
                __init__.py
                app.py
                assets/
                pages/
                    foo_page.py
            rio.toml
        """,
        app_file="my-project/my_project/app.py",
        main_module="my_project.app",
    )
    app = load_user_app(config).app

    assert app.name == "App"
    assert app.assets_dir == Path("my-project/my_project/assets").absolute()
    assert_page_was_loaded_correctly(app.pages[0], "foo-was-loaded-correctly")


@pytest.mark.xfail(reason="Waiting for new pyfakefs version to be published")
def test_import_sibling_module(fs: FakeFilesystem):
    config = create_project(
        fs,
        """
        my-project/
            foo.py
            bar.py
            rio.toml
        """,
        app_file="my-project/foo.py",
        main_module="bar",
    )
    Path("my-project/bar.py").write_text("""
# Note: The import statement doesn't work with pyfakefs, so we have to use
# importlib.
import importlib
app = importlib.import_module('foo').app
""")
    app = load_user_app(config).app

    assert app.name == "Bar"
    assert app.assets_dir == Path("my-project").absolute()
    assert not app.pages


@pytest.mark.xfail(reason="Waiting for new pyfakefs version to be published")
def test_import_from_submodule(fs: FakeFilesystem):
    config = create_project(
        fs,
        """
        my-project/
            my_project/
                __init__.py
                app.py
                assets/
                pages/
                    foo_page.py
            helper.py
            rio.toml
        """,
        app_file="my-project/helper.py",
        main_module="my_project.app",
    )
    Path("my-project/my_project/app.py").write_text("""
# Note: The import statement doesn't work with pyfakefs, so we have to use
# importlib.
import importlib
app = importlib.import_module('helper').app
""")
    app = load_user_app(config).app

    assert app.name == "App"
    assert app.assets_dir == Path("my-project/my_project/assets").absolute()
    assert_page_was_loaded_correctly(app.pages[0], "foo-was-loaded-correctly")


@pytest.mark.xfail(reason="Waiting for new pyfakefs version to be published")
def test_relative_import_from_pages(fs: FakeFilesystem):
    config = create_project(
        fs,
        """
        my-project/
            my_project/
                __init__.py
                app.py
                assets/
                pages/
                    fancy_page.py
            rio.toml
        """,
        app_file="my-project/my_project/app.py",
        main_module="my_project.app",
    )
    Path("my-project/my_project/pages/fancy_page.py").write_text("""
# Note: The import statement doesn't work with pyfakefs, so we have to use
# importlib.
import importlib.util
app_module = importlib.import_module(importlib.util.resolve_name('..app', __package__))
_ = app_module.app  # Just to make sure the right file was imported


import rio

@rio.page(url_segment='fancy-page-was-loaded-correctly')
class FancyPage(rio.Component):
    def build(self):
        return rio.Text('fancy')
""")
    app = load_user_app(config).app

    assert app.name == "App"
    assert app.assets_dir == Path("my-project/my_project/assets").absolute()
    assert_page_was_loaded_correctly(app.pages[0], "foo-was-loaded-correctly")


def test_relative_paths_with_module(fs: FakeFilesystem):
    config = create_project(
        fs,
        """
        my-project/
            my_app.py
            my/
                assets/
                pages/
                    foo_page.py
            rio.toml
        """,
        app_file="my-project/my_app.py",
        main_module="my_app",
    )
    Path("my-project/my_app.py").write_text("""
import rio

app = rio.App(
    build=rio.Spacer,
    assets_dir='my/assets',
    pages='my/pages',
)
""")
    app = load_user_app(config).app

    assert app.name == "My App"
    assert app.assets_dir == Path("my-project/my/assets").absolute()
    assert_page_was_loaded_correctly(app.pages[0], "foo-was-loaded-correctly")


def test_relative_paths_with_package(fs: FakeFilesystem):
    config = create_project(
        fs,
        """
        my-project/
            my_project/
                __init__.py
                my/
                    pages/
                        foo_page.py
            rio.toml
        """,
        app_file="my-project/my_project/__init__.py",
        main_module="my_project",
    )
    Path("my-project/my_project/__init__.py").write_text("""
import rio

app = rio.App(
    build=rio.Spacer,
    assets_dir='my/assets',
    pages='my/pages',
)
""")
    app = load_user_app(config).app

    assert app.name == "My Project"
    assert app.assets_dir == Path("my-project/my_project/my/assets").absolute()
    assert_page_was_loaded_correctly(app.pages[0], "foo-was-loaded-correctly")



"""
Contains documentation related tasks specific to the Rio project.
"""

import collections
import functools
import inspect
import types
import typing as t

import imy.docstrings
import unicall

import rio

__all__ = [
    "get_rio_module_docs",
    "get_all_documented_objects",
    "get_toplevel_documented_objects",
    "get_docs_for",
    "get_documentation_url",
    "get_documentation_url_segment",
    "insert_links_into_markdown",
]


RIO_MODULE_DOCS: imy.docstrings.ModuleDocs | None = None
NAME_TO_DOCS: (
    t.Mapping[
        str,
        t.Sequence[
            imy.docstrings.ClassDocs
            | imy.docstrings.FunctionDocs
            | imy.docstrings.AttributeDocs
            | imy.docstrings.PropertyDocs
            | imy.docstrings.ParameterDocs,
        ],
    ]
    | None
) = None


def _prepare_docs():
    """
    Creates the `imy.docstrings.ModuleDocs` object for the `rio` module. This
    involves some expensive operations, so some useful data structures are
    cached in the global scope.
    """
    global RIO_MODULE_DOCS

    # Let imy parse the module
    RIO_MODULE_DOCS = imy.docstrings.ModuleDocs.from_module(rio)

    # Add stuff that imy doesn't consider public
    RIO_MODULE_DOCS.add_member(rio.event)
    RIO_MODULE_DOCS.add_member(rio.extension_event)

    url_docs = _make_docs_for_rio_url()
    url_docs.owner = RIO_MODULE_DOCS
    RIO_MODULE_DOCS.members["URL"] = url_docs

    # Apply rio-specific post-processing
    postprocess_docs(RIO_MODULE_DOCS)

    # Insert links to other documentation pages
    _insert_hyperlinks_into_rio_docs(RIO_MODULE_DOCS)


def _insert_hyperlinks_into_rio_docs(
    rio_module_docs: imy.docstrings.ModuleDocs,
) -> None:
    # Make a list of all Docs objects we have to process
    all_docs = list(
        docs
        for docs in rio_module_docs.iter_children(
            include_self=True, recursive=True
        )
        # Exclude the getters and setters of properties. We only need the
        # properties themselves.
        if not (isinstance(docs.owner, imy.docstrings.PropertyDocs))
    )

    # Make a mapping from names to Docs objects. A name can refer to multiple
    # objects, e.g. `Rectangle` and `Card` both have a `content` attribute.
    name_to_docs = collections.defaultdict(list)

    for docs in all_docs:
        if isinstance(docs, imy.docstrings.ModuleDocs):
            continue

        names = [docs.name]
        if isinstance(
            docs, (imy.docstrings.ClassDocs, imy.docstrings.FunctionDocs)
        ):
            full_name = docs.full_name
            names += [full_name, full_name.removeprefix("rio.")]

        for name in names:
            name_to_docs[name].append(docs)

    # Cache the mapping, `insert_links_into_markdown` needs it
    global NAME_TO_DOCS
    NAME_TO_DOCS = name_to_docs

    # Loop over all the docs objects and insert hyperlinks into their markdown
    for docs in all_docs:
        # Not everything needs to be hyperlinked. A page doesn't need to link to
        # itself. A parameter doesn't need to link to its function. Etc.
        urls_to_ignore = _get_urls_to_ignore(docs)

        docs.transform_docstrings(
            lambda markdown: insert_links_into_markdown(
                markdown,
                urls_to_ignore=urls_to_ignore,
            )
        )


def _get_urls_to_ignore(docs) -> t.Sequence[rio.URL]:
    if isinstance(docs, imy.docstrings.ModuleDocs):
        return ()

    urls_to_ignore = [url_for_docs(docs)]

    # Class members don't need to link their class
    if isinstance(
        docs, (imy.docstrings.PropertyDocs, imy.docstrings.AttributeDocs)
    ):
        urls_to_ignore += _get_urls_to_ignore(docs.owner)
    elif isinstance(docs, imy.docstrings.FunctionDocs):
        if isinstance(docs.owner, imy.docstrings.ClassDocs):
            urls_to_ignore += _get_urls_to_ignore(docs.owner)
    # Parameters don't need to link to the function
    elif isinstance(docs, imy.docstrings.ParameterDocs):
        urls_to_ignore += _get_urls_to_ignore(docs.owner)

    return urls_to_ignore


def get_rio_module_docs() -> imy.docstrings.ModuleDocs:
    """
    Returns the `imy.docstrings.ModuleDocs` object for the `rio` module.
    """
    if RIO_MODULE_DOCS is None:
        _prepare_docs()
        assert RIO_MODULE_DOCS is not None

    return RIO_MODULE_DOCS


@functools.cache
def get_all_documented_objects() -> (
    dict[
        type | t.Callable | property,
        imy.docstrings.ClassDocs
        | imy.docstrings.FunctionDocs
        | imy.docstrings.PropertyDocs,
    ]
):
    all_docs = get_rio_module_docs().iter_children(
        include_self=True, recursive=True
    )
    return {
        docs.object: docs
        for docs in all_docs
        if isinstance(
            docs,
            (
                imy.docstrings.ClassDocs,
                imy.docstrings.FunctionDocs,
                imy.docstrings.PropertyDocs,
            ),
        )
    }


@functools.cache
def get_toplevel_documented_objects() -> (
    dict[
        type | t.Callable,
        imy.docstrings.ClassDocs | imy.docstrings.FunctionDocs,
    ]
):
    """
    Returns only objects that have their own page in our docs. (That means no
    methods.)
    """
    return {
        docs.object: docs
        for docs in get_all_documented_objects().values()
        if isinstance(
            docs, (imy.docstrings.ClassDocs, imy.docstrings.FunctionDocs)
        )
        and isinstance(docs.owner, imy.docstrings.ModuleDocs)
    }


@t.overload
def get_docs_for(obj: type) -> imy.docstrings.ClassDocs: ...


@t.overload
def get_docs_for(obj: types.FunctionType) -> imy.docstrings.FunctionDocs: ...


def get_docs_for(
    obj: types.FunctionType | t.Type,
) -> imy.docstrings.ClassDocs | imy.docstrings.FunctionDocs:
    """
    Parse the docs for a component and return them. The results are cached, so
    this function is fast.
    """
    return get_all_documented_objects()[obj]  # type: ignore


def _make_docs_for_rio_url():
    docs = imy.docstrings.ClassDocs.from_class(rio.URL)
    docs.attributes.clear()
    docs.members.clear()
    docs.summary = "Alias for `yarl.URL`."
    docs.details = """
Since URLs are a commonly used data type, Rio re-exports `yarl.URL` as
`rio.URL` for convenience. See the
[`yarl` documentation](https://yarl.aio-libs.org/en/stable/api/#yarl.URL) for
details about this class.
""".strip()

    return docs


def get_documentation_url(
    obj: type | t.Callable,
    *,
    relative: bool = False,
) -> rio.URL:
    """
    Returns the URL to the documentation for the given Rio object. This doesn't
    perform any checks on whether the object actually exists and has
    documentation. It relies solely on the passed values.

    (This function is used by the dev tools.)
    """
    # Build the relative URL
    url = f"/docs/api/{get_documentation_url_segment(obj)}"

    # Make it absolute, if requested
    if not relative:
        url = "https://rio.dev" + url

    # Done
    return rio.URL(url)


def get_documentation_url_segment(obj: type | t.Callable) -> str:
    """
    Returns the URL segment for the documentation of the given Rio object. This
    doesn't perform any checks on whether the object actually exists and has
    documentation. It relies solely on the passed in values.

    (This function is used by the rio website to generate pages.)
    """
    return obj.__name__.lower()


def url_for_docs(
    docs: imy.docstrings.ClassDocs
    | imy.docstrings.FunctionDocs
    | imy.docstrings.AttributeDocs
    | imy.docstrings.PropertyDocs
    | imy.docstrings.ParameterDocs,
    *,
    relative: bool = False,
) -> rio.URL:
    # Classes and functions (not methods!) have their own documentation page, so
    # those simply defer to `get_documentation_url`.
    if isinstance(docs, imy.docstrings.ClassDocs):
        return get_documentation_url(docs.object, relative=relative)

    if isinstance(docs, imy.docstrings.FunctionDocs):
        # Methods are listed on the page of the class, so get the url for the
        # class and then add the function name
        if isinstance(docs.owner, imy.docstrings.ClassDocs):
            url = url_for_docs(docs.owner, relative=relative)
            # ScrollTargets currently don't work, so don't create urls with
            # #fragments
            return url  # + f"#{docs.name.lower()}"
        else:
            return get_documentation_url(docs.object)

    # ScrollTargets currently don't work, so don't create urls with #fragments
    """
    # Fields and properties are listed on the page of the class, so get the url
    # for the class and then add an url fragment
    if isinstance(docs, (imy.docstrings.AttributeDocs, imy.docstrings.PropertyDocs)):
        url = url_for_docs(docs.owner, relative=relative)
        return url + f"#{docs.name.lower()}"

    # Parameters are listed on the page of the function, so get the url for the
    # function and then add an url fragment
    if isinstance(docs, imy.docstrings.ParameterDocs):
        url = url_for_docs(docs.owner, relative=relative)

        if "#" in url:
            return url + f".{docs.name.lower()}"
        else:
            return url + f"#{docs.name.lower()}"
    """
    assert docs.owner is not None
    return url_for_docs(docs.owner)

    assert False, f"url_for_docs received invalid input: {docs}"


def postprocess_docs(
    docs: imy.docstrings.FunctionDocs
    | imy.docstrings.ClassDocs
    | imy.docstrings.ModuleDocs,
) -> None:
    """
    Applies rio-specific post-processing.
    """

    # Make the summary into a single line. (This is because the summary is
    # sometimes displayed inside a `rio.Text`, which honors newlines. We
    # don't want that.)
    if docs.summary:
        docs.summary = docs.summary.replace("\n", " ")

    if isinstance(docs, imy.docstrings.ModuleDocs):
        postprocess_module_docs(docs)
    elif isinstance(docs, imy.docstrings.FunctionDocs):
        postprocess_function_docs(docs)
    elif issubclass(docs.object, rio.Component):
        postprocess_component_docs(docs)
    else:
        postprocess_class_docs(docs)


def postprocess_module_docs(docs: imy.docstrings.ModuleDocs) -> None:
    for member in docs.members.values():
        postprocess_docs(member)


def postprocess_function_docs(docs: imy.docstrings.FunctionDocs) -> None:
    pass


def postprocess_class_docs(docs: imy.docstrings.ClassDocs) -> None:
    """
    Perform Rio specific post-processing on the component, such as stripping out
    internal attributes and functions.
    """

    # Strip out anything `Session` inherits from `unicall`
    if docs.name == "Session":
        to_remove = set(dir(unicall.Unicall)).difference(vars(rio.Session))
        docs.members = {
            name: member
            for name, member in docs.members.items()
            if name not in to_remove
        }

    # Strip default docstrings created by dataclasses
    if docs.summary is not None and docs.summary.startswith(f"{docs.name}("):
        docs.summary = None
        docs.details = None

    # Skip internal functions
    def keep_method(func: imy.docstrings.FunctionDocs) -> bool:
        # Internal methods start with an underscore
        keep = not func.name.startswith("_")

        # Some methods in components are meant to be used by the user, but only
        # when they're the one creating the component. For example, the `build`
        # method is only intended to be used by the component itself, and
        # documenting it would be pointless at best, and confusing at worst.
        is_inherited_protected_method = (
            docs.name != "Component"
            and func.name
            in (
                "build",
                "call_event_handler",
                "force_refresh",
            )
        )
        if is_inherited_protected_method:
            keep = False

        # Strip lambdas
        if func.name == "<lambda>":
            keep = False

        # Make sure to keep the constructor
        if func.name == "__init__":
            keep = True

        # Some classes are not meant to be constructed by the user. Strip their
        # constructor.
        if func.name == "__init__":
            if (
                docs.name in ("FileInfo", "Session")
                or docs.name.endswith(("Event", "Error"))
                or inspect.isabstract(docs.object)
            ):
                keep = False

        # Check if it's explicitly excluded
        if not func.metadata.public:
            keep = False

        return keep

    docs.members = {
        name: member
        for name, member in docs.members.items()
        if not isinstance(member, imy.docstrings.FunctionDocs)
        or keep_method(member)
    }

    # Post-process the constructor
    try:
        init_function = docs.members["__init__"]
    except KeyError:
        pass
    else:
        assert isinstance(init_function, imy.docstrings.FunctionDocs)

        # Strip the ridiculous default docstring created by dataclasses
        #
        # FIXME: Not working for some reason
        if (
            init_function.summary
            == "Initialize self. See help(type(self)) for accurate signature."
        ):
            init_function.summary = None
            init_function.details = None

        # Inject a short description for `__init__` if there is none.
        if init_function.summary is None:
            init_function.summary = f"Creates a new `{docs.name}` instance."


def postprocess_component_docs(docs: imy.docstrings.ClassDocs) -> None:
    # Apply the standard class post-processing
    postprocess_class_docs(docs)

    if docs.object is not rio.Component:
        # Remove methods that are only useful in custom components
        docs.members = {
            name: member
            for name, member in docs.members.items()
            if name
            not in ("bind", "build", "call_event_handler", "force_refresh")
        }

        # Subclasses of `rio.Component` inherit a load of constructor
        # parameters, which clutter the docs. We'll sort the keyword-only
        # parameters so that the inherited parameters appear at the end.
        try:
            init_func = docs.members["__init__"]
        except KeyError:
            pass
        else:
            assert isinstance(init_func, imy.docstrings.FunctionDocs)

            parameters = list[imy.docstrings.ParameterDocs]()
            kwargs = list[imy.docstrings.ParameterDocs]()

            for param in init_func.parameters.values():
                if param.kw_only:
                    kwargs.append(param)
                else:
                    parameters.append(param)

            COMPONENT_CONSTRUCTOR_PARAMS = set(rio.Component.__annotations__)
            kwargs.sort(
                key=lambda param: param.name in COMPONENT_CONSTRUCTOR_PARAMS
            )
            parameters += kwargs

            init_func.parameters = {param.name: param for param in parameters}


def insert_links_into_markdown(
    markdown: str,
    *,
    urls_to_ignore: t.Iterable[rio.URL] = (),
) -> str:
    if NAME_TO_DOCS is None:
        _prepare_docs()
        assert NAME_TO_DOCS is not None

    name_to_docs = NAME_TO_DOCS  # Reassign to appease the type checker

    urls_to_ignore_as_str = {str(url) for url in urls_to_ignore}

    def url_for_name(name: str) -> str | None:
        possible_targets = {
            str(url_for_docs(docs)) for docs in name_to_docs.get(name, ())
        }

        # Filter out ignored urls
        possible_targets.difference_update(urls_to_ignore_as_str)

        if not possible_targets:
            return None

        if len(possible_targets) == 1:
            return possible_targets.pop()

        # TODO: Can we somehow figure out which url was meant?
        return None

    return imy.docstrings.insert_links_into_markdown(
        markdown,
        url_for_name=url_for_name,
    )



from __future__ import annotations

import dataclasses
import logging
import typing as t

import imy.docstrings
from uniserde import JsonDoc

import rio

from .. import deprecations, utils
from .fundamental_component import FundamentalComponent

__all__ = [
    "FilePickEvent",
    "FilePickerArea",
]


@t.final
@imy.docstrings.mark_constructor_as_private
@dataclasses.dataclass
class FilePickEvent:
    """
    Holds information regarding a file upload event.

    This is a simple dataclass that stores useful information for when the user
    picks a file using a `FilePickerArea`. You'll typically receive this as
    argument in `on_pick_file` events.

    ## Attributes

    `file`: Handle to the uploaded file.
    """

    file: rio.FileInfo


@t.final
@deprecations.component_kwarg_renamed(
    since="0.10.9",
    old_name="on_choose_file",
    new_name="on_pick_file",
)
class FilePickerArea(FundamentalComponent):
    """
    Drag & Drop are for files

    The `FilePickerArea` component allows the user to upload files either by
    dragging and dropping them onto the component, or optionally using a regular
    file browser. Whenever a file has been uploaded, the `on_pick_file` event
    is triggered, allowing you to run code.

    ## Attributes

    `file_types`: A list of file extensions which the user is allowed
        to select. Defaults to `None`, which means that the user may select any
        file. Values can be passed as file extensions, ('pdf', '.pdf', '*.pdf'
        are all accepted) or MIME types (e.g. 'application/pdf').

        Note that there is no way for Rio to enforce the file type. Users will
        always be able to upload arbitrary files, e.g. by renaming them. Treat
        this as a hint to the user, and so the file browser may filter files,
        but not as a security measure.

    `multiple`: Whether the user is allowed to pick multiple files at once. If
        `False`, a maximum of one file can be picked at a time.

    `files`: A list of files that has been picked by the user. These will be
        displayed to them and they'll also have the ability to remove them,
        triggering the `on_remove_file` event.

    `on_pick_file`: Triggered when the user picks a file, either by dragging
        and dropping it onto the component, or by selecting it in the file
        browser. The event data contains `FileInfo` object, which contains
        information about the uploaded file.

    `on_remove_file`: Triggered when the user removes a file from the list of
        picked files. The event data contains `FileInfo` object, which contains
        information about the removed file.


    ## Metadata

    `experimental`: True
    """

    file_types: list[str] | None = None

    multiple: bool = False

    files: list[rio.FileInfo] = []

    on_pick_file: rio.EventHandler[FilePickEvent] = None
    on_remove_file: rio.EventHandler[FilePickEvent] = None

    # Hide internal fields from the type checker
    if not t.TYPE_CHECKING:
        # At most one of these is set, the other `None`. If both are `None`, the
        # component will display a default message.
        child_text: str | None = dataclasses.field(default=None, init=False)
        child_component: rio.Component | None = dataclasses.field(
            default=None, init=False
        )

    # The serializer can't handle Union types. Override the constructor, so it
    # splits the content into two values
    def __init__(
        self,
        content: str | rio.Component | None = None,
        *,
        file_types: list[str] | None = None,
        multiple: bool = False,
        files: list[rio.FileInfo] = [],
        on_pick_file: rio.EventHandler[FilePickEvent] = None,
        on_remove_file: rio.EventHandler[FilePickEvent] = None,
        key: str | int | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        min_width: float = 0,
        min_height: float = 0,
        # MAX-SIZE-BRANCH max_width: float | None = None,
        # MAX-SIZE-BRANCH max_height: float | None = None,
        grow_x: bool = False,
        grow_y: bool = False,
        align_x: float | None = None,
        align_y: float | None = None,
        # SCROLLING-REWORK scroll_x: t.Literal["never", "auto", "always"] = "never",
        # SCROLLING-REWORK scroll_y: t.Literal["never", "auto", "always"] = "never",
    ) -> None:
        """
        ## Parameters

        `content`: What to display to the user. By default, a simple text message is
            shown. If you pass in a string, the component will display that text
            instead. You may also pass in a component, which will be displayed
            instead of the default look.
        """
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            min_width=min_width,
            min_height=min_height,
            # MAX-SIZE-BRANCH max_width=max_width,
            # MAX-SIZE-BRANCH max_height=max_height,
            grow_x=grow_x,
            grow_y=grow_y,
            align_x=align_x,
            align_y=align_y,
            # SCROLLING-REWORK scroll_x=scroll_x,
            # SCROLLING-REWORK scroll_y=scroll_y,
        )

        if content is None:
            self.child_text = None
            self.child_component = None
        elif isinstance(content, str):
            self.child_text = content
            self.child_component = None
        else:
            self.child_text = None
            self.child_component = content

        self.file_types = file_types
        self.multiple = multiple

        self.files = files

        self.on_pick_file = on_pick_file
        self.on_remove_file = on_remove_file

        self._properties_set_by_creator_.update(
            ("child_text", "child_component")
        )

    def _custom_serialize_(self) -> JsonDoc:
        result: JsonDoc = {
            "files": [
                {
                    # Send the ID as string, because JavaScript numbers have a
                    # much more limited range than Python's
                    "id": str(id(file)),
                    "name": file.name,
                }
                for file in sorted(
                    self.files,
                    key=lambda file: file.name.lower(),
                )
            ],
        }

        if self.file_types is not None:
            result["file_types"] = list(
                {
                    utils.normalize_file_extension(file_type)
                    for file_type in self.file_types
                }
            )

        return result

    async def _on_file_upload_(self, files: list[rio.FileInfo]) -> None:
        """
        Special method that's called by Rio when a file is uploaded directly to
        a component, rather than prior registration. Hand off any received
        files.
        """
        await self._update_files(
            add_files=files,
            remove_file_id=None,
        )

    async def _on_message_(self, msg: t.Any) -> None:
        msg_type = msg.get("type", "<invalid>")

        # When running in a window, having the browser upload a file is rather
        # silly. Instead, it will send a message to the component, which will in
        # turn use the session's file picking abilities.
        if msg_type == "pickFile":
            try:
                # The type checker doesn't quite grok the very advanced concept of
                # booleans. Help it out.
                if self.multiple:
                    picked_files = await self.session.pick_file(
                        file_types=self.file_types,
                        multiple=True,
                    )
                else:
                    picked_files = [
                        await self.session.pick_file(
                            file_types=self.file_types,
                            multiple=False,
                        )
                    ]

            # If no files were selected that's fine. No need to spam stdout.
            except rio.NoFileSelectedError:
                return

            # Update the state & call event handlers
            await self._update_files(
                add_files=picked_files,
                remove_file_id=None,
            )

        # Allow the user to unpick files
        elif msg_type == "removeFile":
            await self._update_files(
                add_files=[],
                remove_file_id=msg.get("fileId"),
            )

        # Wah?!
        else:
            logging.error(
                f"Received invalid message from the frontend. This is either a bug, or somebody is crafting invalid messages on purpose. Received message was: {msg}"
            )

    async def _update_files(
        self,
        *,
        add_files: list[rio.FileInfo],
        remove_file_id: str | None,
    ) -> None:
        """
        Adds a file to the list of picked files, managing state and triggering
        events as needed.
        """
        actually_added_files: list[rio.FileInfo] = []
        actually_removed_files: list[rio.FileInfo] = []

        # Add any new files
        if self.multiple:
            actually_added_files.extend(add_files)
            self.files.extend(add_files)
        elif add_files:
            actually_added_files.append(add_files[0])
            actually_removed_files.extend(self.files)
            self.files.clear()
            self.files.append(add_files[0])

        # Remove any files
        if remove_file_id is not None:
            for ii, file in enumerate(self.files):
                if str(id(file)) == remove_file_id:
                    actually_removed_files.append(self.files.pop(ii))
                    break

        # Call event handlers
        #
        # TODO: Should these be called simultaneously?
        for file in actually_added_files:
            event_data = FilePickEvent(file)
            await self.call_event_handler(
                self.on_pick_file,
                event_data,
            )

        for file in actually_removed_files:
            event_data = FilePickEvent(file)
            await self.call_event_handler(
                self.on_remove_file,
                event_data,
            )

        # Refresh
        if actually_added_files or actually_removed_files:
            self.session._register_dirty_component(
                self,
                include_children_recursively=False,
            )
            await self.session._refresh()


FilePickerArea._unique_id_ = "FilePickerArea-builtin"



from __future__ import annotations

import dataclasses
import typing as t

import imy.docstrings
from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "PointerEvent",
    "PointerMoveEvent",
    "PointerEventListener",
]


@imy.docstrings.mark_constructor_as_private
@dataclasses.dataclass
class PointerEvent:
    """
    Holds information regarding a pointer event.

    This is a simple dataclass that stores useful information for when the user
    interacts with a component using a mouse, touch, or other pointer-style
    device. You'll receive this as argument in a variety of pointer events.


    ## Attributes

    `pointer_type`: What sort of pointer triggered the event. Can be either
        `"mouse"` or `"touch"`.

    `button`: The mouse button that was pressed, if any. For mouse events
        (`pointer_type=="mouse"`), this is either `"left"`, `"middle"`, or
        `"right"`. For other events, this is `None`.

    `window_x`: The x coordinate of the pointer relative to the window. The
        origin is the top-left corner of the window, with larger `x` values
        meaning further to the right.

    `window_y`: The y coordinate of the pointer relative to the window. The
        origin is the top-left corner of the window, with larger `y` values
        meaning further down.

    `component_x`: The x coordinate of the pointer relative to the
        `PointerEventListener` component. The origin is the top-left corner of
        the component, with larger `x` values meaning further to the right.

    `component_y`: The y coordinate of the pointer relative to the
        `PointerEventListener` component. The origin is the top-left corner of
        the component, with larger `y` values meaning further down.
    """

    pointer_type: t.Literal["mouse", "touch"]

    button: t.Literal["left", "middle", "right"] | None

    window_x: float
    window_y: float

    component_x: float
    component_y: float

    @staticmethod
    def _from_message(msg: dict[str, t.Any]) -> PointerEvent:
        return PointerEvent(
            pointer_type=msg["pointerType"],
            button=msg.get("button"),
            window_x=msg["windowX"],
            window_y=msg["windowY"],
            component_x=msg["componentX"],
            component_y=msg["componentY"],
        )


@t.final
@imy.docstrings.mark_constructor_as_private
@dataclasses.dataclass
class PointerMoveEvent(PointerEvent):
    """
    Holds information regarding a pointer move event.

    This is a simple dataclass that stores useful information for when the user
    moves the pointer. You'll typically receive this as argument in
    `on_pointer_move` events.


    ## Attributes

    `relative_x`: How far the pointer has moved horizontally since the last
        time the event was triggered.

    `relative_y`: How far the pointer has moved vertically since the last time
        the event was triggered.
    """

    relative_x: float
    relative_y: float

    @staticmethod
    def _from_message(msg: dict[str, t.Any]) -> PointerMoveEvent:
        return PointerMoveEvent(
            pointer_type=msg["pointerType"],
            button=msg.get("button"),
            window_x=msg["windowX"],
            window_y=msg["windowY"],
            component_x=msg["componentX"],
            component_y=msg["componentY"],
            relative_x=msg["relativeX"],
            relative_y=msg["relativeY"],
        )


@t.final
class PointerEventListener(FundamentalComponent):
    """
    Allows you to listen for mouse & touch events on a component.

    `PointerEventListener` takes a single child component and displays it. It
    then listens for any mouse and touch activity on the child component and
    reports it through its events.


    ## Attributes

    `content`: The child component to display and watch.

    `on_press`: Similar to `on_pointer_up`, but performs additional subtle
        checks, such as that the pressed mouse button was the left one.

    `on_pointer_down`: Triggered when a pointer button is pressed down while
        the pointer is placed over the child component.

    `on_pointer_up`: Triggered when a pointer button is released while the
        pointer is placed over the child component.

    `on_pointer_move`: Triggered when the pointer is moved while located over
        the child component.

    `on_pointer_enter`: Triggered when the pointer previously was not located
        over the child component, but is now.

    `on_pointer_leave`: Triggered when the pointer previously was located over
        the child component, but is no longer.

    `on_drag_start`: Triggered when the user starts dragging the pointer, i.e.
        moving it while holding down a pointer button.

    `on_drag_move`: Triggered when the user moves the pointer while holding down
        a pointer button. Note that once a drag event was triggered on a
        component, the move event will continue to fire even if the pointer
        leaves the component.

    `on_drag_end`: Triggered when the user stops dragging the pointer.
    """

    content: rio.Component
    _: dataclasses.KW_ONLY
    on_press: rio.EventHandler[PointerEvent] = None
    on_pointer_down: rio.EventHandler[PointerEvent] = None
    on_pointer_up: rio.EventHandler[PointerEvent] = None
    on_pointer_move: rio.EventHandler[PointerMoveEvent] = None
    on_pointer_enter: rio.EventHandler[PointerEvent] = None
    on_pointer_leave: rio.EventHandler[PointerEvent] = None
    on_drag_start: rio.EventHandler[PointerEvent] = None
    on_drag_move: rio.EventHandler[PointerMoveEvent] = None
    on_drag_end: rio.EventHandler[PointerEvent] = None

    def _custom_serialize_(self) -> JsonDoc:
        return {
            "reportPress": self.on_press is not None,
            "reportPointerDown": self.on_pointer_down is not None,
            "reportPointerUp": self.on_pointer_up is not None,
            "reportPointerMove": self.on_pointer_move is not None,
            "reportPointerEnter": self.on_pointer_enter is not None,
            "reportPointerLeave": self.on_pointer_leave is not None,
            "reportDragStart": self.on_drag_start is not None,
            "reportDragMove": self.on_drag_move is not None,
            "reportDragEnd": self.on_drag_end is not None,
        }

    async def _on_message_(self, msg: t.Any) -> None:
        # Parse the message
        assert isinstance(msg, dict), msg

        msg_type = msg["type"]
        assert isinstance(msg_type, str), msg_type

        # Dispatch the correct event
        if msg_type == "press":
            await self.call_event_handler(
                self.on_press,
                PointerEvent._from_message(msg),
            )

        elif msg_type == "pointerDown":
            await self.call_event_handler(
                self.on_pointer_down,
                PointerEvent._from_message(msg),
            )

        elif msg_type == "pointerUp":
            await self.call_event_handler(
                self.on_pointer_up,
                PointerEvent._from_message(msg),
            )

        elif msg_type == "pointerMove":
            await self.call_event_handler(
                self.on_pointer_move,
                PointerMoveEvent._from_message(msg),
            )

        elif msg_type == "pointerEnter":
            await self.call_event_handler(
                self.on_pointer_enter,
                PointerEvent._from_message(msg),
            )

        elif msg_type == "pointerLeave":
            await self.call_event_handler(
                self.on_pointer_leave,
                PointerEvent._from_message(msg),
            )

        elif msg_type == "dragStart":
            await self.call_event_handler(
                self.on_drag_start,
                PointerEvent._from_message(msg),
            )

        elif msg_type == "dragMove":
            await self.call_event_handler(
                self.on_drag_move,
                PointerMoveEvent._from_message(msg),
            )

        elif msg_type == "dragEnd":
            await self.call_event_handler(
                self.on_drag_end,
                PointerEvent._from_message(msg),
            )

        else:
            raise ValueError(
                f"{__class__.__name__} encountered unknown message: {msg}"
            )

        # Refresh the session
        await self.session._refresh()


PointerEventListener._unique_id_ = "PointerEventListener-builtin"



from __future__ import annotations

# <additional-imports>
import rio

from ... import components as comps
from ... import data_models, theme

# </additional-imports>

# <component>
# After creating our blog page that returns a `rio.PageView` and its
# corresponding sub-page directory, we can now set up a blog section to display
# the blog posts. Since `blog_page.py` has the `url_segment` "blog", we can
# designate the blog home page as the parent of the blog posts by setting its
# `url_segment` to an empty string. This makes the blog home page the parent
# page of all blog posts. Therefore, the url of the blog home page will be
# "website.com/blog". The blog posts will have urls like
# "website.com/blog/post1".


@rio.page(
    name="blog",
    url_segment="",
)
class BlogPage(rio.Component):
    """
    The blog home page.

    This page serves as the main hub for all blog posts. It displays an
    introductory section and a list of blog posts. As the parent page, it
    organizes and manages the URLs for individual blog entries.
    """

    def _desktop_build(self) -> rio.Component:
        """
        Build the layout optimized for desktop devices.
        """
        return rio.Column(
            rio.Text(
                "Blog",
                style=theme.BOLD_BIGGER_SECTION_TITEL_DESKTOP,
            ),
            rio.Text(
                "Welcome to our blog. Here you can find all the latest news and updates.",
                style=theme.DARK_TEXT_BIGGER,
            ),
            rio.Separator(
                color=self.session.theme.neutral_color,
                margin_top=3,
            ),
            comps.BlogSection(),
            spacing=1,  # Space between components for desktop view
            align_x=0.5,  # Center horizontally
            align_y=0,  # Align to the top vertically
            min_width=80,  # Fixed width to ensure proper layout on larger screens
        )

    def _mobile_build(self) -> rio.Component:
        """
        Build the layout optimized for mobile devices.
        """
        return rio.Column(
            rio.Text(
                "Blog",
                overflow="wrap",
                style=theme.BOLD_BIGGER_SECTION_TITEL_MOBILE,
            ),
            rio.Text(
                "Welcome to our blog. Here you can find all the latest news and updates.",
                overflow="wrap",
                style=theme.DARK_TEXT_BIGGER,
            ),
            rio.Separator(
                color=theme.SEPARATOR_COLOR,
                margin_top=3,
            ),
            # Add the BlogSection component to display the blog posts
            comps.BlogSection(),
            align_y=0,  # Align to the top vertically
            spacing=1,  # Space between components for mobile view
            margin_x=1,  # Horizontal margin for mobile screens
        )

    def build(self) -> rio.Component:
        """
        Construct and return the page layout based on the user's device type.

        Determines whether the user is on a desktop or mobile device and builds
        the corresponding layout.
        """
        device = self.session[data_models.PageLayout].device

        if device == "desktop":
            # Build and return the desktop-optimized layout
            return self._desktop_build()

        else:
            # Build and return the mobile-optimized layout
            return self._mobile_build()


# </component>



import dataclasses
import typing as t

import typing_extensions as te


@dataclasses.dataclass(frozen=True)
class Version:
    major: int
    minor: int = 0
    patch: int = 0
    rc: int | None = None

    @classmethod
    def parse(cls, version_string: str) -> te.Self:
        """
        Quick & dirty version parser, useful for comparing versions.

        The results are the major, minor and patch version numbers. The final value
        is negative if the version is a pre-release version and zero otherwise.

        This function is best-effort and only supports version schemes as currently
        used by Rio. It will fail with a `ValueError` if the version string is
        invalid, or simply not supported. Guard against that!
        """
        head, _, tail = version_string.partition("rc")

        if tail:
            rc = int(tail)
        else:
            rc = None

        parts = head.split(".")
        parts += ("0", "0", "0")

        return cls(
            major=int(parts[0]),
            minor=int(parts[1]),
            patch=int(parts[2]),
            rc=rc,
        )

    def bump_major(self) -> te.Self:
        return type(self)(major=self.major + 1)

    def bump_minor(self) -> te.Self:
        return type(self)(major=self.major, minor=self.minor + 1)

    def bump_patch(self) -> te.Self:
        return type(self)(
            major=self.major,
            minor=self.minor,
            patch=self.patch + 1,
        )

    def bump_rc(self) -> te.Self:
        return type(self)(
            major=self.major,
            minor=self.minor,
            patch=self.patch,
            rc=0 if self.rc is None else self.rc + 1,
        )

    def drop_rc(self) -> te.Self:
        return type(self)(
            major=self.major,
            minor=self.minor,
            patch=self.patch,
            rc=None,
        )

    def __gt__(self, other: t.Any) -> bool:
        if not isinstance(other, Version):
            return NotImplemented

        if self.major != other.major:
            return self.major > other.major

        if self.minor != other.minor:
            return self.minor > other.minor

        if self.patch != other.patch:
            return self.patch > other.patch

        if self.rc is None:
            return other.rc is not None

        if other.rc is None:
            return False

        return self.rc > other.rc

    def __str__(self) -> str:
        version_string = f"{self.major}.{self.minor}"

        if self.patch > 0:
            version_string += f".{self.patch}"

        if self.rc is not None:
            version_string += f"rc{self.rc}"

        return version_string



import typing as t

# <additional-imports>
import rio

from .. import components as comps
from .. import theme

# </additional-imports>


# <component>
GOODIES = [
    "GB Storage",
    "Email Account",
    "Domain",
    "Website",
    "Database",
    "SSL Certificate",
    "Support Ticket",
]


class PricingCards(rio.Component):
    """
    A component that displays pricing information for different subscription
    plans.


    ## Attributes:

    `plan_type`: The type of the plan.

    `plan_description`: A description of the plan's features.

    `duration`: The billing cycle of the plan (e.g., "Monthly" or "Yearly").

    `monthly_price`: The price for a monthly subscription.

    `yearly_price`: The price for a yearly subscription.

    `amount_goodies`: The number of included goodies in the plan.

    `is_selected`: Whether the plan is currently selected by the user.
    """

    plan_type: t.Literal["Basic", "Standard", "Premium"]
    plan_description: str
    duration: str
    monthly_price: float
    yearly_price: float
    amount_goodies: int

    is_selected: bool = False

    def build(self) -> rio.Component:
        # Determine pricing details based on the subscription duration
        if self.duration == "Monthly":
            duration = "month"
            price = self.monthly_price
        else:
            duration = "year"
            price = self.yearly_price

        # Define the button based on the selection status
        if self.is_selected:
            button = rio.Button(
                rio.Text(
                    "Get Started",
                    style=rio.TextStyle(
                        fill=self.session.theme.background_color
                    ),
                    align_x=0.5,
                ),
                color=rio.Color.from_hex("ffffff"),
                margin_y=1,
            )
        else:
            button = comps.OutlinedNeutralButton("Get Started", margin_y=1)

        # Create a column of goodies based on the amount of goodies included in
        # the plan
        goody_column = rio.Column(spacing=1)
        # Loop through the list of goodies and add them to the column
        for goodie in GOODIES:
            goody_column.add(
                comps.GoodyColumn(
                    amount_goodies=self.amount_goodies,
                    goodie_text=goodie,
                )
            )

        # Construct the card content
        content = rio.Column(
            rio.Text(
                self.plan_type,
                style=rio.TextStyle(
                    font_weight="bold",
                    font_size=1.8,
                    fill=theme.TEXT_FILL_BRIGHTER,
                ),
            ),
            rio.Text(
                self.plan_description,
                style=theme.DARKER_TEXT,
            ),
            rio.Row(
                rio.Text(
                    f"${price}",
                    style=rio.TextStyle(
                        font_weight="bold",
                        font_size=1.8,
                        fill=theme.TEXT_FILL_BRIGHTER,
                    ),
                ),
                rio.Text(
                    f" /{duration}",
                    style=theme.DARKER_TEXT,
                    align_y=0.85,
                ),
                align_x=0,  # Align the price and duration horizontally
                margin_top=1,  # Add margin above the price section
            ),
            # Add the button to the card
            button,
            # Add the column of goodies to the card
            goody_column,
            spacing=1,  # Add spacing between card elements
            margin=3,  # Add outer margin for the card content
        )

        # Wrap the content in a rectangle with styling based on the selection
        return rio.Rectangle(
            content=content,
            fill=rio.Color.TRANSPARENT,
            corner_radius=self.session.theme.corner_radius_medium,
            # Highlight the border if selected
            stroke_color=(
                self.session.theme.primary_color
                if self.is_selected
                else theme.NEUTRAL_COLOR_BRIGHTER
            ),
            # Adjust border thickness if selected
            stroke_width=0.15 if self.is_selected else 0.1,
        )


# </component>



import threading
import typing as t

T = t.TypeVar("T")


class ThreadsafeFuture(t.Generic[T]):
    def __init__(self) -> None:
        self._result_value: t.Any
        self._event = threading.Event()

    def set_result(self, result: T) -> None:
        self._result_value = result
        self._event.set()

    def result(self) -> T:
        self._event.wait()
        return self._result_value

    def done(self) -> bool:
        return self._event.is_set()



import asyncio
import inspect
import typing as t

import rio.testing


class TracingExtensionParent(rio.Extension):
    def __init__(self) -> None:
        # Keeps track of called functions
        self.function_call_log: list[str] = []

    def verify_and_clear_log(self, *expected: str) -> None:
        """
        Verifies that the recorded function calls match the expected ones. This
        disregards order!
        """

        assert set(self.function_call_log) == set(expected)
        self.function_call_log.clear()

    def _record_function_call(self) -> None:
        # Get the caller's name
        caller_name = inspect.stack()[1].function

        # Record it
        self.function_call_log.append(caller_name)

    # This function should be inherited and called
    @rio.extension_event.on_session_start
    def on_session_start_parent(
        self,
        event: rio.ExtensionSessionStartEvent,
    ) -> None:
        self._record_function_call()


class TracingExtensionChild(TracingExtensionParent):
    @rio.extension_event.on_app_start
    def on_app_start(
        self,
        event: rio.ExtensionAppStartEvent,
    ) -> None:
        self._record_function_call()

    @rio.extension_event.on_app_close
    def on_app_close(
        self,
        event: rio.ExtensionAppCloseEvent,
    ) -> None:
        self._record_function_call()

    # This function isn't registered at all and should not be called
    def on_session_start(
        self,
        event: rio.ExtensionSessionStartEvent,
    ) -> None:
        self._record_function_call()

    # This function is asynchronous, testing that the extension system awaits
    # functions as needed.
    @rio.extension_event.on_session_start
    def on_session_start_async(
        self,
        event: rio.ExtensionSessionStartEvent,
    ) -> None:
        self._record_function_call()

    @rio.extension_event.on_session_close
    def on_session_close(
        self,
        event: rio.ExtensionSessionCloseEvent,
    ) -> None:
        self._record_function_call()

    @rio.extension_event.on_page_change
    def on_page_change(
        self,
        event: rio.ExtensionPageChangeEvent,
    ) -> None:
        self._record_function_call()

    # This function is registered for multiple events and should be called for
    # each of them
    @rio.extension_event.on_session_start
    @rio.extension_event.on_session_close
    @rio.extension_event.on_page_change
    def on_multiple_events(
        self,
        event: t.Any,
    ) -> None:
        self._record_function_call()


async def test_extension_events() -> None:
    # Create an app
    app = rio.App(
        build=lambda: rio.Text("Hello"),
    )

    # Add an extension which records called events
    extension_instance = TracingExtensionChild()
    app._add_extension(extension_instance)

    assert len(extension_instance.function_call_log) == 0

    async with rio.testing.TestClient(app) as test_client:
        # The test client doesn't support Rio's full feature set and so doesn't
        # actually call the app start/close events right now. Skip them

        # Expect: Session start events
        extension_instance.verify_and_clear_log(
            "on_session_start_async",
            "on_session_start_parent",
            "on_multiple_events",
        )

        # Navigate to another page
        test_client.session.navigate_to("/foobar")

        # Expect: Page change event
        #
        # Rio dispatches this event in an asyncio task. Give it some time to
        # run.
        await asyncio.sleep(0.1)

        extension_instance.verify_and_clear_log(
            "on_page_change",
            "on_multiple_events",
        )

    # Expect: Session close events
    extension_instance.verify_and_clear_log(
        "on_session_close",
        "on_multiple_events",
    )

    # The app close event is missing here for the same reason as above



import rio

# <additional-imports>
from .. import data_models, theme

# </additional-imports>


# <component>
class BulletPoint(rio.Component):
    """
    A component that displays a bullet point with an icon, title, and
    description.


    ## Attributes:

    `title`: The title text of the bullet point.

    `description`: The description text of the bullet point.

    `icon`: The icon displayed next to the bullet point.
    """

    title: str
    description: str
    icon: str

    def build(self) -> rio.Component:
        """
        Builds the bullet point component.

        This component consists of:
        - An icon aligned with the title and description.
        - A title with bold styling.
        - A description that adapts to the device layout for mobile and desktop.
        """
        device = self.session[data_models.PageLayout].device
        return rio.Row(
            rio.Icon(
                self.icon,
                fill="primary",
                min_height=1.5,
                min_width=1.5,
                align_y=0,
            ),
            rio.Column(
                # Title text of the bullet point
                rio.Text(
                    self.title,
                    style=theme.BOLD_BRIGHTER_TEXT,
                ),
                # Description text
                rio.Text(
                    self.description,
                    # Adapt text wrapping to device type
                    overflow="wrap" if device == "mobile" else "nowrap",
                    # Ensure minimum width for mobile devices
                    min_width=18 if device == "mobile" else 0,  # TODO
                ),
            ),
            spacing=1,
            margin_y=0.5,
            align_x=0,
        )


# </component>



import pytest

import rio


@pytest.mark.parametrize(
    "laba_in,laba_out",
    [],
)
def create_from_oklab(
    laba_in: tuple[float, float, float, float],
    laba_out: tuple[float, float, float, float],
) -> None:
    """
    Instantiates a `Color` object from oklab values and verifies the internally
    stored values match the expected ones.
    """
    color = rio.Color.from_oklab(*laba_in)

    assert color._l == laba_out[0]
    assert color._a == laba_out[1]
    assert color._b == laba_out[2]
    assert color._opacity == laba_out[3]



from __future__ import annotations

import typing as t

from . import class_container

__all__ = [
    "Spacer",
]


@t.final
class Spacer(class_container.ClassContainer):
    """
    An invisible component which grows by default.

    Spacers are invisible components which add empty space between other
    components. While similar effects can often be achieved using margins and
    alignment, code utilizing spacers can sometimes be easier to read.

    Note that unlike most components in Rio, `Spacer` does not have a `natural`
    size. Instead it defaults to a width and height of `grow`, as that is how
    they're most frequently used.


    ## Examples

    This example will display two texts in a Row, with one of them being pushed
    to the very left and the other to the very right:

    ```python
    rio.Row(
        rio.Text("Hello"),
        rio.Spacer(),
        rio.Text("World"),
    )
    ```
    """

    def __init__(
        self,
        *,
        min_width: float = 0,
        min_height: float = 0,
        # MAX-SIZE-BRANCH max_width: float | None = None,
        # MAX-SIZE-BRANCH max_height: float | None = None,
        grow_x: bool = True,
        grow_y: bool = True,
        key: str | int | None = None,
    ) -> None:
        super().__init__(
            None,
            ["rio-spacer"],
            key=key,
            min_width=min_width,
            min_height=min_height,
            # MAX-SIZE-BRANCH max_width=max_width,
            # MAX-SIZE-BRANCH max_height=max_height,
            grow_x=grow_x,
            grow_y=grow_y,
        )

    def _get_debug_details_(self) -> dict[str, t.Any]:
        result = super()._get_debug_details_()

        # Don't inherit the content from `rio.ClassContainer`.
        del result["content"]

        return result


# Make sure the component is recognized as `ClassContainer`, rather than a new
# component.
Spacer._unique_id_ = class_container.ClassContainer._unique_id_



from __future__ import annotations

import dataclasses
import typing as t

import imy.docstrings
from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "Drawer",
    "DrawerOpenOrCloseEvent",
]


@t.final
@imy.docstrings.mark_constructor_as_private
@dataclasses.dataclass
class DrawerOpenOrCloseEvent:
    """
    Holds information regarding a drawer open or close event.

    This is a simple dataclass that stores useful information for when the user
    opens or closes a drawer. You'll typically received this as argument in
    `on_open_or_close` events.

    ## Attributes

    `is_open`: The new `is_open` state of the `Drawer`.
    """

    is_open: bool


@t.final
class Drawer(FundamentalComponent):
    """
    A container which slides in from the edge of the screen.

    Drawers are containers which can either be completely hidden from view, or
    be made visible by sliding in from the edge of the screen. They are commonly
    used for navigation on smaller displays.

    Drawers take two children: The `anchor` is always visible and positions the
    drawer. The `content` is located inside the drawer and is only visible when
    the drawer is open.

    Drawers have the ability to be `modal`. Modal drawers draw attention to
    themselves and prevent interaction with the anchor while open.


    ## Attributes

    `anchor`: A component which is always visible and positions the drawer.

    `content`: A component which is only visible when the drawer is open.

    `on_open_or_close`: Triggered whenever the user opens or closes the
        drawer.

    `side`: The side of the screen from which the drawer slides in.

    `is_modal`: Whether the drawer should prevent interaction with the anchor
        while open.

    `is_open`: Whether the drawer is currently open.

    `is_user_openable`: Whether the user can open or close the drawer. If this
        is `False`, the drawer can only be opened or closed
        programmatically.

    `color`: The color scheme to use for the drawer content. The drawer itself
        will use the specified color, while content will automatically use one
        that is legible on top of it. This has no effect on the anchor.


    ## Examples

    A simple drawer with a button as the anchor and some text as content:

    ```python
    rio.Drawer(
        anchor=rio.Button("Click Me!"),
        content=rio.Text("It was clickbait!"),
    )
    ```

    Same as above, but wired so the drawer opens when the button is clicked:

    ```python
    class MyComponent(rio.Component):
        is_open: bool = False

        def on_press_button(self) -> None:
            self.is_open = True

        def build(self) -> rio.Component:
            return rio.Drawer(
                anchor=rio.Button(
                    "Click Me!",
                    on_press=self.on_press_button,
                ),
                content=rio.Text(
                    "It was clickbait!",
                ),
                is_open=self.is_open,
            )
    ```
    """

    anchor: rio.Component
    content: rio.Component
    _: dataclasses.KW_ONLY
    on_open_or_close: rio.EventHandler[DrawerOpenOrCloseEvent] = None
    side: t.Literal["left", "right", "top", "bottom"] = "left"
    is_modal: bool = True
    is_open: bool = False
    is_user_openable: bool = True
    color: rio.ColorSet = "neutral"

    def _validate_delta_state_from_frontend(self, delta_state: JsonDoc) -> None:
        if not set(delta_state) <= {"is_open"}:
            raise AssertionError(
                f"Frontend tried to change `{type(self).__name__}` state: {delta_state}"
            )

        if "is_open" in delta_state and not self.is_user_openable:
            raise AssertionError(
                "Frontend tried to change value of `Drawer.is_open` even though `is_user_openable` is `False`"
            )

    async def _call_event_handlers_for_delta_state(
        self, delta_state: JsonDoc
    ) -> None:
        # Trigger on_open_or_close event
        try:
            is_open = delta_state["is_open"]
        except KeyError:
            pass
        else:
            assert isinstance(is_open, bool), (is_open, type(is_open))
            await self.call_event_handler(
                self.on_open_or_close,
                DrawerOpenOrCloseEvent(is_open),
            )


Drawer._unique_id_ = "Drawer-builtin"



from __future__ import annotations

import dataclasses
import typing as t
from pathlib import Path

import rio
import rio.docs

try:
    RIO_PATH = Path(rio.__file__).parent
except Exception:
    # Use a dummy path. It doesn't really matter, as long as it doesn't crash
    # `Path.is_relative_to`
    RIO_PATH = Path.cwd() / "rio"


class ComponentAttributes(rio.Component):
    component_id: int

    _: dataclasses.KW_ONLY

    on_switch_to_layout_view: rio.EventHandler[[]] = None

    component_natural_width: float = 0
    component_natural_height: float = 0

    component_allocated_inner_width: float = 0
    component_allocated_inner_height: float = 0

    @rio.event.on_populate
    async def fetch_client_side_details(self) -> None:
        """
        Fetch additional details about the component which are only known
        client-side. Stores them in the class.
        """
        # Remember which component the details are being fetched for
        component_id = self.component_id

        # Fetch the details
        try:
            (layout,) = await self.session._get_component_layouts(
                [component_id]
            )
        except KeyError:
            return

        # If the current component has changed while the values were fetched,
        # don't update the state
        if self.component_id != component_id:
            return

        # Publish the results
        self.component_natural_width = layout.natural_width
        self.component_natural_height = layout.natural_height
        self.component_allocated_inner_width = layout.allocated_inner_width
        self.component_allocated_inner_height = layout.allocated_inner_height

    def build(self) -> rio.Component:
        # Get the target component
        try:
            target = self.session._weak_components_by_id[self.component_id]

        # If the component can't be found, don't display anything. This can
        # happen, e.g. with injected components or due to network lag.
        except KeyError:
            return rio.Spacer(min_height=0)

        # Get the debug details. This is a dictionary which contains all
        # keys/values the component has, excluding internals.
        debug_details = target._get_debug_details_()

        # Build the result. There is no need to add a heading, because it was
        # already handled by the parent component.
        result = DetailsGrid()

        # Which file/line was this component instantiated from?
        file, line = target._creator_stackframe_

        # If it was instantiated somewhere in the rio internals, don't show it
        if not file.is_relative_to(RIO_PATH):
            try:
                file = file.relative_to(Path.cwd())
            except ValueError:
                pass

            result.add_full_width(
                rio.Markdown(
                    f"Created in `{file}`, line {line}",
                    margin_top=0.5,
                    margin_bottom=1,
                ),
            )

        # The component's attributes
        self._build_details(result, target, debug_details)

        # Push all of the content to the left. This could be done by aligning
        # the entire Grid, but that would ellipsize some long texts. Instead,
        # add a Spacer into a fifth column, which will take up any unused space.
        result.add(
            rio.Spacer(grow_y=False),
            column=4,
        )

        # Push the remaining content to the bottom
        result.add_full_width(rio.Spacer())

        # Link to docs
        if type(target)._rio_builtin_:
            docs_url = rio.URL(rio.docs.get_documentation_url(type(target)))

            result.add_full_width(
                rio.Link(
                    "Read the Docs",
                    icon="material/library_books",
                    target_url=docs_url,
                    open_in_new_tab=True,
                    margin_top=1,
                    align_x=0,
                )
            )

        # Offer to show the detailed layout subpage
        result.add_full_width(
            rio.Button(
                "Layout View",
                icon="material/space_dashboard",
                on_press=self.on_switch_to_layout_view,
                shape="rounded",
            )
        )

        # Done!
        return result.as_rio_component()

    def _build_details(
        self,
        result: DetailsGrid,
        target: rio.Component,
        debug_details: dict[str, t.Any],
    ) -> None:
        # Add the component's attributes
        result.add_heading3("Custom Attributes", margin_top=0)
        has_custom_attributes: bool = False

        for prop_name, prop_value in debug_details.items():
            # Make sure to skip any which already have custom tailored cells
            if prop_name in (
                "min_width",
                "min_height",
                "grow_x",
                "grow_y",
                "margin",
                "margin_x",
                "margin_y",
                "margin_left",
                "margin_right",
                "margin_top",
                "margin_bottom",
                "align_x",
                "align_y",
            ):
                continue

            # Display this property
            result.add_row(prop_name, repr(prop_value))
            has_custom_attributes = True

        if not has_custom_attributes:
            result.add_full_width(
                rio.Text(
                    "This component has no custom attributes",
                    justify="left",
                    style="dim",
                    margin=0.6,
                )
            )

        # Layout stuff: size, grow, alignment, margin
        result.add_heading3("Layout")

        # Size
        if "min_width" in debug_details or "min_height" in debug_details:
            try:
                py_min_width = debug_details["min_width"]
            except KeyError:
                py_min_width_str = "-"
            else:
                if isinstance(py_min_width, (int, float)):
                    py_min_width = round(py_min_width, 2)

                py_min_width_str = repr(py_min_width)

            try:
                py_min_height = debug_details["min_height"]
            except KeyError:
                py_min_height_str = "-"
            else:
                if isinstance(py_min_height, (int, float)):
                    py_min_height = round(py_min_height, 2)

                py_min_height_str = repr(py_min_height)

            # Add some extra spacing
            result.row += 1

            # Header
            result.add_label("width", column=1, justify="right")
            result.add_label("height", column=2, justify="right")
            result.row += 1

            # The minimum size as specified in Python
            result.add_label("min", column=0)
            result.add_value(py_min_width_str, column=1, justify="right")
            result.add_value(py_min_height_str, column=2, justify="right")
            result.row += 1

            # The component's natural size
            result.add_label("natural", column=0)
            result.add_value(
                str(round(self.component_natural_width, 2)),
                column=1,
                justify="right",
            )
            result.add_value(
                str(round(self.component_natural_height, 2)),
                column=2,
                justify="right",
            )
            result.row += 1

            # The component's allocated size
            result.add_label("allocated", column=0)
            result.add_value(
                str(round(self.component_allocated_inner_width, 2)),
                column=1,
                justify="right",
            )
            result.add_value(
                str(round(self.component_allocated_inner_height, 2)),
                column=2,
                justify="right",
            )
            result.row += 1

        # Grow
        if "grow_x" in debug_details or "grow_y" in debug_details:
            result.add_label("grow_x", column=0)
            result.add_value(str(debug_details.get("grow_x", "-")), column=1)

            result.add_label("grow_y", column=2)
            result.add_value(str(debug_details.get("grow_y", "-")), column=3)

            result.row += 1

        # Alignment
        if "align_x" in debug_details or "align_y" in debug_details:
            result.add_label("align_x", column=0)
            result.add_value(str(debug_details.get("align_x", "-")), column=1)

            result.add_label("align_y", column=2)
            result.add_value(str(debug_details.get("align_y", "-")), column=3)

            result.row += 1

        # Margins
        margin_left = target._effective_margin_left_
        margin_top = target._effective_margin_top_
        margin_right = target._effective_margin_right_
        margin_bottom = target._effective_margin_bottom_

        single_x_margin = margin_left == margin_right
        single_y_margin = margin_top == margin_bottom

        if single_x_margin and single_y_margin:
            result.add_label("margin", column=0)
            result.add_value(str(margin_left), column=1)

            result.row += 1

        else:
            if single_x_margin:
                result.add_label("margin_x", column=0)
                result.add_value(str(margin_left), column=1)

            else:
                result.add_label("margin_left", column=0)
                result.add_value(str(margin_left), column=1)

                result.add_label("margin_right", column=2)
                result.add_value(str(margin_right), column=3)

            result.row += 1

            if single_y_margin:
                result.add_label("margin_y", column=0)
                result.add_value(str(margin_top), column=1)

            else:
                result.add_label("margin_top", column=0)
                result.add_value(str(margin_top), column=1)

                result.add_label("margin_bottom", column=2)
                result.add_value(str(margin_bottom), column=3)

            result.row += 1


class DetailsGrid:
    def __init__(self, **kwargs) -> None:
        self.grid = rio.Grid(row_spacing=0.5, column_spacing=0.5, **kwargs)
        self.row = 0

    def add_row(self, label: str, value: str) -> None:
        self.add_label(
            label,
            column=0,
            component_width=8,
            ellipsize=True,
        )
        self.add_value(
            value,
            column=1,
            width=3,
            ellipsize=True,
        )
        self.row += 1

    def add_label(
        self,
        text: str,
        *,
        row: int | None = None,
        column: int,
        justify: t.Literal["left", "center", "right"] = "right",
        ellipsize: bool = False,
        component_width: float = 0,
        column_width: int = 1,
    ) -> None:
        self.add(
            rio.Text(
                text,
                style="dim",
                overflow="ellipsize" if ellipsize else "nowrap",
                justify=justify,
                min_width=component_width,
            ),
            row=row,
            column=column,
            width=column_width,
        )

    def add_value(
        self,
        value: str,
        *,
        row: int | None = None,
        column: int,
        width: int = 1,
        justify: t.Literal["left", "center", "right"] = "left",
        ellipsize: bool = False,
    ) -> None:
        self.add(
            rio.Text(
                value,
                justify=justify,
                grow_x=ellipsize,
                overflow="ellipsize" if ellipsize else "nowrap",
            ),
            row=row,
            column=column,
            width=width,
        )

    def add(
        self,
        child: rio.Component,
        *,
        row: int | None = None,
        column: int,
        width: int = 1,
        height: int = 1,
    ) -> None:
        if row is None:
            row = self.row

        self.grid.add(child, row, column, width=width, height=height)

    def add_full_width(self, child: rio.Component) -> None:
        self.grid.add(
            child,
            self.row,
            0,
            width=5,
        )
        self.row += 1

    def add_heading3(
        self,
        heading: str,
        *,
        margin_top=1,
    ) -> None:
        self.add_full_width(
            rio.Text(
                text=heading,
                style="heading3",
                justify="left",
                margin_top=margin_top,
            )
        )

    def add_spacing(self, amount: float = 1) -> None:
        self.add_full_width(rio.Spacer(min_height=amount))

    def as_rio_component(self) -> rio.Component:
        return self.grid



from __future__ import annotations

import typing as t

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "Tooltip",
]


@t.final
class Tooltip(FundamentalComponent):
    """
    A small pop-up window that appears when the user hovers over an element.

    `Tooltip` is a small pop-up window that appears when the user hovers over an
    element. It is commonly used to provide additional information about the
    element, such as a description or a hint.


    ## Attributes

    `anchor`: The component to which the tooltip is anchored.

    `tip`: The text or component to display in the tooltip.

    `position`: The position of the tooltip relative to the anchor. It can be
        one of the following values: `left`, `top`, `right`, `bottom`.

    `gap`: The distance between the tooltip and its anchor.


    ## Examples

    This example will display a label for an icon when the user hovers over it:

    ```python
    rio.Tooltip(
        anchor=rio.Icon("material/info"),
        tip="This is a tooltip.",
        position="top",
    )
    ```
    """

    anchor: rio.Component
    tip: str | rio.Component
    position: t.Literal["auto", "left", "top", "right", "bottom"]
    gap: float

    # Hide internal attributes from the IDE
    if not t.TYPE_CHECKING:
        _tip_component: rio.Component | None

    # Impute a Text instance if a string is passed in as the tip
    def __init__(
        self,
        anchor: rio.Component,
        tip: str | rio.Component,
        position: t.Literal["auto", "left", "top", "right", "bottom"] = "auto",
        *,
        gap: float = 0.5,
        key: str | int | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        min_width: float = 0,
        min_height: float = 0,
        # MAX-SIZE-BRANCH max_width: float | None = None,
        # MAX-SIZE-BRANCH max_height: float | None = None,
        grow_x: bool = False,
        grow_y: bool = False,
        align_x: float | None = None,
        align_y: float | None = None,
        # SCROLLING-REWORK scroll_x: t.Literal["never", "auto", "always"] = "never",
        # SCROLLING-REWORK scroll_y: t.Literal["never", "auto", "always"] = "never",
    ) -> None:
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            min_width=min_width,
            min_height=min_height,
            # MAX-SIZE-BRANCH max_width=max_width,
            # MAX-SIZE-BRANCH max_height=max_height,
            grow_x=grow_x,
            grow_y=grow_y,
            align_x=align_x,
            align_y=align_y,
            # SCROLLING-REWORK scroll_x=scroll_x,
            # SCROLLING-REWORK scroll_y=scroll_y,
        )

        self.anchor = anchor
        self.tip = tip

        if isinstance(tip, str):
            self._tip_text = tip
            self._tip_component = None
        else:
            self._tip_text = None
            self._tip_component = tip

        self.position = position
        self.gap = gap

        self._properties_set_by_creator_.add("_tip_component")

    def __post_init__(self) -> None:
        # FIXME: This breaks attribute bindings
        if isinstance(self._tip_text, str):
            self._tip_component = rio.Text(
                self._tip_text,
                selectable=False,
            )


Tooltip._unique_id_ = "Tooltip-builtin"



from .author_card import AuthorCard as AuthorCard
from .blog_header import BlogHeader as BlogHeader
from .blog_major import BlogMajor as BlogMajor
from .blog_minor import BlogMinor as BlogMinor
from .blog_section import BlogSection as BlogSection
from .bullet_point import BulletPoint as BulletPoint
from .card_section import CardSection as CardSection
from .custom_switcher_bar import CustomSwitcherBar as CustomSwitcherBar
from .fake_link import FakeLink as FakeLink
from .faq import Faq as Faq
from .faq_section import FaqSection as FaqSection
from .footer import Footer as Footer
from .get_started import GetStarted as GetStarted
from .goody_column import GoodyColumn as GoodyColumn
from .hero_card import HeroCard as HeroCard
from .hero_section import HeroSection as HeroSection
from .image_placeholder import ImagePlaceholder as ImagePlaceholder
from .major_column import MajorColumn as MajorColumn
from .navbar import Navbar as Navbar
from .outlined_colored_button import (
    OutlinedColoredButton as OutlinedColoredButton,
)
from .outlined_neutral_button import (
    OutlinedNeutralButton as OutlinedNeutralButton,
)
from .pricing_cards import PricingCards as PricingCards
from .pricing_section import PricingSection as PricingSection
from .root_component import RootComponent as RootComponent
from .service_power import ServicePower as ServicePower
from .service_speed import ServiceSpeed as ServiceSpeed
from .subscribe_input_button import SubscribeInputButton as SubscribeInputButton
from .testimonial import Testimonial as Testimonial
from .testimonials import Testimonials as Testimonials



# <additional-imports>
import rio

from .. import components as comps
from .. import data_models

# </additional-imports>


# <component>
@rio.page(
    name="Pricing",
    url_segment="pricing",
)
class PricingPage(rio.Component):
    """
    The Pricing page of the website.

    This page provides detailed pricing information for Services, including
    various pricing plans and frequently asked questions (FAQs) to assist users
    in making informed decisions.

    ## Attributes:

    `duration`: The duration of the pricing plan.
    """

    duration: str = "Monthly"

    def _desktop_build(self) -> rio.Component:
        """
        Build the layout optimized for desktop devices.
        """
        return rio.Column(
            comps.PricingSection(),
            comps.FaqSection(),
            spacing=5,  # Consistent spaceing for desktop readability
            align_x=0.5,  # Center horizontally
            align_y=0,  # Align to the top vertically
            min_width=80,  # Fixed width to ensure proper layout on larger screens
        )

    def _mobile_build(self) -> rio.Component:
        """
        Build the layout optimized for mobile devices.
        """
        return rio.Column(
            comps.PricingSection(),
            comps.FaqSection(),
            spacing=5,  # Consistent spacing for mobile readability
            align_x=0.5,  # Center horizontally
            align_y=0,  # Align to the top vertically
            margin_x=1,  # Horizontal margin for compact layout on mobile
        )

    def build(self) -> rio.Component:
        """
        Construct and return the page layout based on the user's device type.

        Determines whether the user is on a desktop or mobile device and builds
        the corresponding layout.
        """
        device = self.session[data_models.PageLayout].device

        if device == "desktop":
            # Build and return the desktop-optimized layout
            return self._desktop_build()

        else:
            # Build and return the mobile-optimized layout
            return self._mobile_build()


# </component>



import functools
import html
import sys
import types
import typing as t
from pathlib import Path

import path_imports

import rio
import rio.app_server.fastapi_server
import rio.global_state

from ... import icon_registry, nice_traceback, project_config


class AppLoadError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)

    @property
    def message(self) -> str:
        return self.args[0]


def make_traceback_html(
    *,
    err: t.Union[str, BaseException],
    project_directory: Path,
) -> str:
    error_icon_svg = icon_registry.get_icon_svg("material/error")

    if isinstance(err, str):
        traceback_html = html.escape(err)
    else:
        traceback_html = nice_traceback.format_exception_html(
            err,
            relpath=project_directory,
        )

    return f"""
<div class="rio-traceback rio-dev-tools-background rio-switcheroo-neutral">
    <div class="rio-traceback-header">
        {error_icon_svg}
        <div>Couldn't load the app</div>
    </div>
    <div class="rio-traceback-message">
        Fix the issue below. The app will automatically reload once you save the
        file.
    </div>
    <div class="rio-traceback-traceback">{traceback_html}</div>
    <div class="rio-traceback-footer">
        Need help?
        <div class="rio-traceback-footer-links">
            <a class="rio-text-link" target="_blank" href="https://discord.gg/7ejXaPwhyH">Ask on Rio's Discord</a>
            <a class="rio-text-link" target="_blank" href="https://chatgpt.com">Ask ChatGPT</a>
            <a class="rio-text-link" target="_blank" href="https://rio.dev/docs?s=bf2">Read the docs</a>
        </div>
    </div>
</div>
"""


def make_error_message_component(
    err: t.Union[str, BaseException],
    project_directory: Path,
) -> rio.Component:
    html = make_traceback_html(
        err=err,
        project_directory=project_directory,
    )

    return rio.Webview(
        html,
        align_x=0.5,
        align_y=0.2,
    )


def make_error_message_app(
    err: t.Union[str, BaseException],
    project_directory: Path,
    theme: rio.Theme | tuple[rio.Theme, rio.Theme],
) -> rio.App:
    """
    Creates an app that displays the given error message.
    """
    return rio.App(
        build=functools.partial(
            make_error_message_component, err, project_directory
        ),
        # Make sure to explicitly suppress any pages. If not passed, Rio will
        # attempt to auto-detect the pages, which aren't meant for the
        # placeholder app.
        pages=[],
        theme=theme,
    )


def modules_in_directory(project_path: Path) -> t.Iterable[str]:
    """
    Returns all currently loaded modules that reside in the given directory (or
    any subdirectory thereof). As a second condition, modules located inside of
    a virtual environment are not returned.

    The purpose of this is to yield all modules that belong to the user's
    project. This is the set of modules that makes sense to reload when the user
    makes a change.
    """
    # Resolve the path to avoid issues with symlinks
    project_path = project_path.resolve()

    # Walk all modules
    for name, module in list(sys.modules.items()):
        # Special case: Unloading Rio, while Rio is running is not that smart.
        #
        # Also make sure "__main__" isn't unloaded. This can happen because the
        # project is running as `rio run` (which makes rio "__main__") and if
        # rio is located in the project directory.
        if name in ("__main__", "rio") or name.startswith("rio."):
            continue

        # Where does the module live?
        #
        # This needs to take namespace packages into account, which don't have a
        # single path.
        module_paths: t.Iterable[object]

        try:
            module_paths = iter(module.__path__)
        except KeyError:
            # `__path__` is a weird iterable that executes code when you iterate
            # over it. This code crashes with a KeyError if the parent module
            # is missing from `sys.modules`.
            #
            # In such a case, the safe play is to reload the module. The
            # worst case scenario is that we'll reload a module that didn't need
            # to be reloaded, which is much better than *not* reloading a module
            # that should be.
            yield name
            continue
        except AttributeError:
            module_paths = [getattr(module, "__file__", None)]
        except TypeError:
            continue

        # If any path is inside of the process directory, yield the module
        for module_path in module_paths:
            # Get the path
            try:
                module_path = Path(module_path).resolve()  # type: ignore
            except TypeError:
                continue

            # If the module isn't inside of the project directory, skip it
            try:
                module_rel_path = module_path.relative_to(project_path)
            except ValueError:
                continue

            # If the module is unlikely to be related to this project, skip it
            if "site-packages" in module_rel_path.parts:
                continue

            # Yield the module
            yield name
            break


def import_app_module(
    proj: project_config.RioProjectConfig,
) -> types.ModuleType:
    """
    This function imports the app module, as specified by the user.

    The module will be freshly imported, even if it was already imported before.
    """
    # Purge all modules that belong to the project. While the main module name
    # is known, deleting only that isn't enough in all projects. In complex
    # project structures it can be useful to have the UI code live in a module
    # that is then just loaded into a top-level Python file.
    #
    # `modules_in_directory` returns an iterable that can still run code while
    # being iterated over. Some of that code attempts to access modules from
    # `sys.modules`. If we were to start deleting modules while the iterable is
    # still being generated, some of this code can fail because it can't find
    # its module. To avoid this, store everything in a list first, then walk
    # that list again.
    modules = list(modules_in_directory(proj.project_directory))

    for module_name in modules:
        del sys.modules[module_name]

    # Explicitly tell the app what the "main file" is, because otherwise it
    # would be detected incorrectly.
    app_main_module_path = proj.find_app_main_module_path()
    rio.global_state.rio_run_app_module_path = app_main_module_path

    # Now (re-)import the app module. There is no need to import all the other
    # modules here, since they'll be re-imported as needed by the app module.
    try:
        mod = path_imports.import_from_path(
            app_main_module_path,
            proj.app_main_module_name,
            import_parent_modules=True,
            # Newbies often don't organize their code as a single module, so to
            # guarantee that all their files can be imported, we'll add the
            # relevant directory to `sys.path`
            add_parent_directory_to_sys_path=True,
        )
        return mod

    finally:
        rio.global_state.rio_run_app_module_path = None


def load_user_app(
    proj: project_config.RioProjectConfig,
) -> rio.app_server.fastapi_server.FastapiServer:
    """
    Load and return the user's app. Raises `AppLoadError` if the app can't be
    loaded for whichever reason.

    The result is actually an app server, rather than just the app. This is done
    so the user can use `as_fastapi` on a Rio app and still use `rio run` to run
    it. If you actually need the app object itself, grab it from within the
    server.
    """
    # Import the app module
    try:
        app_module = import_app_module(proj)
    except FileNotFoundError as err:
        raise AppLoadError(
            f"Could not import `{proj.app_main_module_name}`: Module not found"
        )
    except ImportError as err:
        real_error = err.__cause__
        assert real_error is not None, err

        raise AppLoadError(
            f"Could not import `{proj.app_main_module_name}`"
        ) from real_error

    # Find the variable holding the Rio app.
    #
    # There are two cases here. Typically, there will be an instance of
    # `rio.App` somewhere. However, in order for users to be able to add custom
    # routes, there might also be a variable storing a `fastapi.FastAPI`, or, in
    # our case, Rio's subclass thereof. If that is present, prefer it over the
    # plain Rio app.
    as_fastapi_apps: list[
        tuple[str, rio.app_server.fastapi_server.FastapiServer]
    ] = []
    rio_apps: list[tuple[str, rio.App]] = []

    for var_name, var in app_module.__dict__.items():
        if isinstance(var, rio.app_server.fastapi_server.FastapiServer):
            as_fastapi_apps.append((var_name, var))

        elif isinstance(var, rio.App):
            rio_apps.append((var_name, var))

    # Prepare the main file name
    if app_module.__file__ is None:
        main_file_reference = f"Your app's main file"
    else:
        main_file_reference = f"The file `{Path(app_module.__file__).relative_to(proj.project_directory)}`"

    # Which type of app do we have?
    #
    # Case: FastAPI app
    if len(as_fastapi_apps) > 0:
        app_list = as_fastapi_apps
        app_server = as_fastapi_apps[0][1]
    # Case: Rio app
    elif len(rio_apps) > 0:
        app_list = rio_apps
        app_instance = rio_apps[0][1]
        app_server = app_instance.as_fastapi()
        assert isinstance(
            app_server, rio.app_server.fastapi_server.FastapiServer
        )
    # Case: No app
    else:
        raise AppLoadError(
            f"Cannot find your app. {main_file_reference} needs to to define a"
            f" variable that is a Rio app. Something like `app = rio.App(...)`"
        )

    # Make sure there was only one app to choose from, within the chosen
    # category
    if len(app_list) > 1:
        variables_string = (
            "`" + "`, `".join(var_name for var_name, _ in app_list) + "`"
        )
        raise AppLoadError(
            f"{main_file_reference} defines multiple Rio apps: {variables_string}. Please make sure there is exactly one."
        )

    return app_server



from __future__ import annotations

import dataclasses
import re
import typing as t
from pathlib import Path

__all__ = [
    "Snippet",
    "SnippetManager",
]


SECTION_PATTERN = re.compile(r" *#\s*<(\/?[\w-]+)>")


@dataclasses.dataclass
class Snippet:
    # The group the snippet is in
    group: str

    # The name the snippet can be accessed with
    name: str

    # The path to the snippet file
    file_path: Path

    # For text snippets, this is the raw file contents. `None` for binary
    # snippets.
    raw_code: str | None

    @property
    def is_text_snippet(self) -> bool:
        return self.raw_code is not None

    @property
    def is_binary_snippet(self) -> bool:
        return not self.is_text_snippet

    @staticmethod
    def from_path(group: str, name: str, file_path: Path) -> Snippet:
        # Read the contents if it's a text snippet
        if file_path.suffix in (".txt", ".md", ".py", ".json"):
            raw_code = file_path.read_text(encoding="utf-8")

            # The example projects contain a bunch of imports that can't be
            # resolved, so those are marked with special `# type: ignore`
            # comments that need to be removed
            if file_path.suffix == ".py":
                raw_code = raw_code.replace(
                    "  # type: ignore (hidden from user)", ""
                )
        else:
            raw_code = None

        return Snippet(
            group=group,
            name=name,
            file_path=file_path,
            raw_code=raw_code,
        )

    def get_section(self, section_name: str) -> str:
        assert self.is_text_snippet, self
        assert self.raw_code is not None, self  # For the type checker

        # Find the target section
        lines = self.raw_code.splitlines()
        section_found = False
        result = []

        for line in lines:
            match = SECTION_PATTERN.match(line)

            # No match
            if match is None:
                result.append(line)
                continue

            # Start of the target section?
            if match.group(1) == section_name:
                result = []
                section_found = True
                continue

            # End of the target section?
            if match.group(1) == f"/{section_name}":
                return "\n".join(result)

            # Some other section, drop it
            pass

        # The section was never found
        if not section_found:
            raise KeyError(f"There is no section named `{section_name}`")

        # The section was found, but never closed
        raise ValueError(f"The section `{section_name}` was never closed")

    def stripped_code(self) -> str:
        """
        Returns the given code with all section tags removed.
        """
        assert self.is_text_snippet, self
        assert self.raw_code is not None, self  # For the type checker

        lines = self.raw_code.splitlines()

        # Remove all section tags
        result = []
        for line in lines:
            if SECTION_PATTERN.match(line) is None:
                result.append(line)

        return "\n".join(result)


class SnippetManager:
    """
    A class the helps load snippets from a directory, caching the results.
    """

    def __init__(self, snippet_directory: Path) -> None:
        # The directory containing all snippet groups
        self._snippet_directory = snippet_directory

        # Once initialized, stores all snippets by group and name. Instead of
        # accessing this value directly, call `_get_snippet_cache()`. That
        # function ensures that the cache has been initialized and returns it.
        self._snippet_cache: dict[str, dict[str, Snippet]] | None = None

    def _get_all_snippet_paths(self) -> dict[str, dict[str, Path]]:
        """
        Returns a dictionary containing all available snippets. The snippets are
        organized by group and name.
        """
        # Find all snippet files
        result = {}

        def scan_dir_recursively(group_name: str, path: Path) -> None:
            assert path.is_dir(), path
            assert result is not None

            # Exclude some directories
            if path.name == "__pycache__":
                return

            for fpath in path.iterdir():
                # Directory
                if fpath.is_dir():
                    scan_dir_recursively(group_name, fpath)

                # Snippet file
                else:
                    name = fpath.name
                    group = result.setdefault(group_name, {})
                    group[name] = fpath

        # Scan all snippet directories. The first directory is used as a key, the
        # rest just for organization.
        for group_dir in self._snippet_directory.iterdir():
            assert group_dir.is_dir(), group_dir
            scan_dir_recursively(group_dir.name, group_dir)

        return result

    def _get_snippet_cache(self) -> dict[str, dict[str, Snippet]]:
        """
        Ensures that the snippet cache has been initialized. If it hasn't
        snippets are loaded from disk first, then returned.

        This will always return the same dictionary instance. Thus the result
        must not be modified.
        """
        # If the cache has already been initialized, return it
        if self._snippet_cache is not None:
            return self._snippet_cache

        # Load all snippets
        self._snippet_cache = {}

        for group_name, snippets in self._get_all_snippet_paths().items():
            group_dict = self._snippet_cache[group_name] = {}

            for name, file_path in snippets.items():
                group_dict[name] = Snippet.from_path(
                    group_name,
                    name,
                    file_path,
                )

        return self._snippet_cache

    def get_snippet(self, group: str, name: str) -> Snippet:
        """
        Gets the snippet with the given group and name.

        This function always returns the same snippet instance for a given
        group and name. Thus the result must not be modified.

        ## Raises

        `KeyError`: If there is no snippet with the given group and name.
        """
        all_groups = self._get_snippet_cache()

        group_dict = all_groups[group]
        return group_dict[name]

    def get_all_snippet_groups(self) -> set[str]:
        """
        Returns a set of all available snippet groups.
        """
        all_groups = self._get_snippet_cache()
        return set(all_groups.keys())

    def get_all_snippets_in_group(self, group: str) -> t.Iterable[Snippet]:
        """
        Returns all snippets in the given group.

        This function always returns the same snippet instances for a given
        group. Thus the result must not be modified.

        ## Raises

        `KeyError`: If there is no group with the given name.
        """
        all_groups = self._get_snippet_cache()
        group_dict = all_groups[group]
        return tuple(group_dict.values())



import io
import tempfile
import typing as t

import pytest

import rio.utils

ASCII_BYTES = (b"Hello, World!", "Hello, World!")
UTF8_BYTES = (b"\xe2\x98\x83", "☃")  # Snowman
LATIN1_BYTES = (b"\xa9", "©")  # Copyright symbol
BLOB_BYTES = (b"\x80\x80", None)  # Invalid ASCII and UTF-8


def create_blobs_variants(blob: bytes) -> t.Iterable[bytes | t.IO[bytes]]:
    """
    Given a bytes object, return it in several variants:

    - As a bytes object
    - As in-memory stream (io.BytesIO)
    - As a real file (tempfile.NamedTemporaryFile)
    """

    # In-memory bytes object
    yield blob

    # A stream
    yield io.BytesIO(blob)

    # A real friggin file
    f = tempfile.NamedTemporaryFile(delete=False)
    f.write(blob)
    f.seek(0)
    yield f


def make_test_blobs() -> (
    t.Iterable[
        tuple[
            bytes,
            bytes | t.IO[bytes],
        ]
    ]
):
    """
    Generate a variety of blobs to use as file contents, in different forms.
    """

    blobs = [
        ASCII_BYTES[0],
        UTF8_BYTES[0],
        LATIN1_BYTES[0],
        BLOB_BYTES[0],
    ]

    for as_bytes in blobs:
        for as_some_blob in create_blobs_variants(as_bytes):
            yield as_bytes, as_some_blob


def make_test_texts() -> (
    t.Iterable[
        t.Tuple[
            bytes | t.IO[bytes],
            str | None,
            str | t.Type[Exception],
        ],
    ]
):
    """
    Generate a variety of blobs to be decoded into strings, in different formats.
    Each entry contains the blob, the encoding to use, and the expected output.
    """
    as_bytes = [
        (
            ASCII_BYTES[0],
            "ascii",
            ASCII_BYTES[1],
        ),
        (
            ASCII_BYTES[0],
            "utf-8",
            ASCII_BYTES[1],
        ),
        (
            UTF8_BYTES[0],
            "ascii",
            UnicodeDecodeError,
        ),
        (
            UTF8_BYTES[0],
            "utf-8",
            UTF8_BYTES[1],
        ),
        (
            LATIN1_BYTES[0],
            "ascii",
            UnicodeDecodeError,
        ),
        (
            LATIN1_BYTES[0],
            "utf-8",
            UnicodeDecodeError,
        ),
        (
            LATIN1_BYTES[0],
            "latin1",
            LATIN1_BYTES[1],
        ),
        (
            BLOB_BYTES[0],
            "ascii",
            UnicodeDecodeError,
        ),
        (
            BLOB_BYTES[0],
            "utf-8",
            UnicodeDecodeError,
        ),
    ]

    # Expand the blobs into multiple formats
    for as_bytes, encoding, expected in as_bytes:
        for as_some_blob in create_blobs_variants(as_bytes):
            yield as_some_blob, encoding, expected


@pytest.mark.parametrize(
    "as_bytes,as_some_blob",
    make_test_blobs(),
)
async def test_read_bytes(
    as_bytes: bytes,
    as_some_blob: bytes | t.IO[bytes],
) -> None:
    """
    Create a `FileInfo`, read back the bytes and make sure the output matches
    the input.
    """

    # Create a FileInfo object
    file_info = rio.utils.FileInfo(
        name="name",
        size_in_bytes=0,
        media_type="application/octet-stream",
        contents=as_some_blob,
    )

    # Read the contents back
    contents_out = await file_info.read_bytes()

    assert as_bytes == contents_out


@pytest.mark.parametrize(
    "as_some_blob,encoding,expected",
    make_test_texts(),
)
async def test_read_text(
    as_some_blob: bytes | t.IO[bytes],
    encoding: str,
    expected: t.Type[Exception] | str,
) -> None:
    """
    Create a `FileInfo`, read back the text and make sure the output matches
    the input, or raises the appropriate exception.
    """

    # Create a FileInfo object
    file_info = rio.utils.FileInfo(
        name="name",
        size_in_bytes=0,
        media_type="application/octet-stream",
        contents=as_some_blob,
    )

    # Try to read the contents back
    if isinstance(expected, str):
        contents_out = await file_info.read_text(encoding=encoding)
        assert contents_out == expected

    else:
        with pytest.raises(expected):
            await file_info.read_text(encoding=encoding)


@pytest.mark.parametrize(
    "as_bytes,as_some_blob",
    make_test_blobs(),
)
async def test_open_bytes(
    as_bytes: bytes,
    as_some_blob: bytes | t.IO[bytes],
) -> None:
    """
    Create a `FileInfo`, open it as a binary file and make sure the output
    matches the input.
    """

    # Create a FileInfo object
    file_info = rio.utils.FileInfo(
        name="name",
        size_in_bytes=0,
        media_type="application/octet-stream",
        contents=as_some_blob,
    )

    # Open the file
    as_file = await file_info.open("rb")
    contents_out = as_file.read()

    assert as_bytes == contents_out


@pytest.mark.parametrize(
    "as_some_blob,encoding,expected",
    make_test_texts(),
)
async def test_open_text(
    as_some_blob: bytes | t.IO[bytes],
    encoding: str,
    expected: t.Type[Exception] | str,
) -> None:
    """
    Create a `FileInfo`, open it as a text file and make sure the output matches
    the input.
    """

    # Create a FileInfo object
    file_info = rio.utils.FileInfo(
        name="name",
        size_in_bytes=0,
        media_type="application/octet-stream",
        contents=as_some_blob,
    )

    # Open the file
    if isinstance(expected, str):
        as_file = await file_info.open("r", encoding=encoding)
        contents_out = as_file.read()
        assert contents_out == expected

    else:
        with pytest.raises(expected):
            as_file = await file_info.open("r", encoding=encoding)
            as_file.read()
